package net.sf.corn.gate;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * This config implementation reads xml files that suits GateSchema.xsd and 
 * registers the contained information to the Registry.
 * @author Serhat Dirik
 */
public class XMLConfig {

	private static Logger log = LoggerFactory
			.getLogger(XMLConfig.class);
	private static List<Document> XMLDocs = new ArrayList<Document>();
	private static Schema schema;
	private static Validator validator;
	private static String namespaceURI = "http://net.sf.corn.gate/schema/config";
	private static ReentrantLock lock = new ReentrantLock();

	static {
		// Create a SchemaFactory capable of understanding WXS schemas.
		SchemaFactory factory = SchemaFactory
				.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		ClassLoader cl = XMLConfig.class.getClassLoader();
		// Load a WXS schema, represented by a Schema instance.
		InputStream xsdStream = cl
				.getResourceAsStream("net/sf/corn/gate/GateSchema.xsd");
		Source schemaFile = new StreamSource(xsdStream);
		try {
			schema = factory.newSchema(schemaFile);
			// Create a Validator object, which can be used to validate
			// an instance document.
			validator = schema.newValidator();
		} catch (SAXException e) {
			log.error(e.getMessage(), e);
		}

	}

	/**
	 * Registers an XML config file to process later on with
	 * "processRegisteredXMLConfig" method.
	 * 
	 * @param configXml
	 */
	public static void registerXMLConfig(InputStream configXml) {
		if (configXml == null) {
			throw new IllegalArgumentException("Input stream is null!");
		}
		lockState();
		// XMLDoc = null;
		log.info("XMLConfigHelper reading stream  ...");
		try {
			DocumentBuilder lOBDocumentBuilder = getDocumentBuilder();
			// Step 1: parse the input file
			Document XMLDoc = lOBDocumentBuilder.parse(configXml);
			if (XMLDoc == null) {
				log.error("Couldn't parse the config xml!");
				return;
			}
			// Step 2 : validate the file
			validator.validate(new DOMSource(XMLDoc));
			// Step 3 : add to process later
			XMLDocs.add(XMLDoc);
		} catch (SAXException se) {
			log.error("Error validating Registry Config file:"
					+ se.getLocalizedMessage());
		} catch (Exception ioe) {
			log.error("An error occured while registering xml configuration",
					ioe);
		}
		releaseState();

	}

	public static void clearXMLCache() {
		lockState();
		XMLDocs.clear();
		releaseState();
	}

	/**
	 * Processes registered config files, resolves them and registers services,
	 * locators, filters, paths within those files
	 */
	public static void processRegisteredXMLConfig() {
		log.info("XMLConfigHelper parsing config files...");
		lockState();
		try {
			// Register locators
			for (Document XMLDoc : XMLDocs) {
				registerLocatorsFromConfig(XMLDoc);
			}
			// Register packages
			for (Document XMLDoc : XMLDocs) {
				registerPackagesFromConfig(XMLDoc);
			}
			// Register filters
			for (Document XMLDoc : XMLDocs) {
				registerFiltersFromConfig(XMLDoc);
			}

			// Register Services
			for (Document XMLDoc : XMLDocs) {
				registerServicesFromConfig(XMLDoc);
			}
			// Register Context Wrapper
			for (Document XMLDoc : XMLDocs) {
				registerContextWrapperFromConfig(XMLDoc);
			}

		} catch (Exception ex) {
			log.error("Registration Error!", ex);
		} finally {
			releaseState();
		}
		XMLDocs = new ArrayList<Document>();
	}

	/**
	 * parse and register locators from config
	 */
	private static void registerPackagesFromConfig(Document XMLDoc) {
		log.info("XMLConfigHelper registering packages...");
		// Find Locators
		NodeList nodes = XMLDoc.getElementsByTagNameNS(namespaceURI, "Package");

		List<String> includes = new ArrayList<String>();
		List<String> excludes = new ArrayList<String>();

		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			NamedNodeMap attributes = node.getAttributes();
			if (attributes == null) {
				log.error("Invalid package definition detected !");
				continue;
			}
			// Attributes

			String packageName = null;
			boolean exclude = false;

			for (int j = 0; j < attributes.getLength(); j++) {
				Node attribute = attributes.item(j);
				if (attribute == null || attribute.getNodeValue() == null
						|| attribute.getNodeValue().trim().length() == 0) {
					continue;
				}
				if (attribute.getLocalName().equals("pattern")) {
					packageName = attribute.getNodeValue();
				} else if (attribute.getLocalName().equals("exclude")) {
					exclude = Boolean.valueOf(attribute.getNodeValue());
				}
			}
			if (packageName == null) {
				log.error("Invalid package definition detected !");
				continue;
			}
			if (exclude)
				excludes.add(packageName);
			else
				includes.add(packageName);
		}

		if (includes.size() > 0) {
			try {
				Registry.scanAndRegister(includes, excludes);
			} catch (Exception e) {
				log.error(
						"An error occured while registering a packages to scan!",
						e);
			}

		}

	}

	/**
	 * parse and register locators from config
	 */
	private static void registerLocatorsFromConfig(Document XMLDoc) {
		log.info("XMLConfigHelper registering locators...");
		// Find Locators
		NodeList nodes = XMLDoc.getElementsByTagNameNS(namespaceURI, "Locator");
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			NamedNodeMap attributes = node.getAttributes();
			if (attributes == null) {
				log.error("Invalid locator definition detected !");
				continue;
			}
			// Attributes

			String clazz = null;

			for (int j = 0; j < attributes.getLength(); j++) {
				Node attribute = attributes.item(j);
				if (attribute == null || attribute.getNodeValue() == null
						|| attribute.getNodeValue().trim().length() == 0) {
					continue;
				}
				if (attribute.getLocalName().equals("className")) {
					clazz = attribute.getNodeValue();
					try {
						Registry.registerLocator(clazz);
					} catch (GateException e) {
						log.error("An error occured while registering a locator!"
								+ " Locator  class="
								+ clazz
								+ ". Skipping registration! ");
					}
					break;
				}
			}
		}

	}

	/**
	 * parse and register context wrapper from config
	 */
	private static void registerContextWrapperFromConfig(Document XMLDoc) {
		log.info("XMLConfigHelper registering context wrapper...");
		// Find Locators
		NodeList nodes = XMLDoc.getElementsByTagNameNS(namespaceURI,
				"ContextWrapper");
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			NamedNodeMap attributes = node.getAttributes();
			if (attributes == null) {
				log.error("Invalid Context Wrapper definition detected !");
				continue;
			}
			// Attributes

			String clazz = null;

			for (int j = 0; j < attributes.getLength(); j++) {
				Node attribute = attributes.item(j);
				if (attribute == null || attribute.getNodeValue() == null
						|| attribute.getNodeValue().trim().length() == 0) {
					continue;
				}
				if (attribute.getLocalName().equals("className")) {
					clazz = attribute.getNodeValue();
					try {
						Registry.registerContextWrapper(clazz);
					} catch (GateException e) {
						log.error("An error occured while registering the context wrapper!"
								+ " Context Wrapper  class "
								+ clazz
								+ ". Skipping registration! ");
					}
					break;
				}
			}
		}

	}

	/**
	 * Registers Interceptors from config
	 */
	private static void registerFiltersFromConfig(Document XMLDoc) {
		log.info("XMLConfigHelper registering interceptor chains...");
		// Find Interceptors
		NodeList nodes = XMLDoc.getElementsByTagNameNS(namespaceURI, "Filter");
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			NamedNodeMap attributes = node.getAttributes();
			if (attributes == null) {
				log.error("Invalid filter definition detected !");
				continue;
			}

			String name = null, description = null;

			for (int j = 0; j < attributes.getLength(); j++) {
				Node attribute = attributes.item(j);
				if (attribute == null || attribute.getNodeValue() == null
						|| attribute.getNodeValue().trim().length() == 0) {
					continue;
				}
				if (attribute.getLocalName().equals("name")) {
					name = attribute.getNodeValue();
				} else if (attribute.getLocalName().equals("description")) {
					description = attribute.getNodeValue();
				}
			}

			ServiceFilter def = null;

			try {
				def = new ServiceFilter(name);
				def.description(description);
			} catch (Exception ex) {
				log.error("An error occured while registering Filter!"
						+ " Filter Name=" + name + ". Skipping registration! ");
				continue;
			}

			NodeList childNodes = node.getChildNodes();
			for (int k = 0; k < childNodes.getLength(); k++) {
				Node aChildNode = childNodes.item(k);
				if (aChildNode.getLocalName() == null
						|| aChildNode.getLocalName().trim().equals("#text")
						|| aChildNode.getLocalName().trim().equals("#comment")) {
					continue;
				}
				NamedNodeMap childAttributes = aChildNode.getAttributes();
				if (aChildNode.getLocalName().trim().equals("ServiceTarget")) {
					String pattern = null;
					for (int j = 0; j < childAttributes.getLength(); j++) {
						Node childAttribute = childAttributes.item(j);
						if (childAttribute == null
								|| childAttribute.getNodeValue() == null
								|| childAttribute.getNodeValue().trim()
										.length() == 0) {
							continue;
						}
						String val = childAttribute.getNodeValue().trim();
						if (childAttribute.getLocalName().equals("pattern")) {
							pattern = val;
						}
					}

					ServiceTargetDefinition st = new ServiceTargetDefinition(
							pattern);
					NodeList subChildNodes = aChildNode.getChildNodes();
					for (int l = 0; l < subChildNodes.getLength(); l++) {
						Node subchildNode = subChildNodes.item(l);
						if (subchildNode.getLocalName() == null
								|| subchildNode.getLocalName().trim()
										.equals("#text")
								|| subchildNode.getLocalName().trim()
										.equals("#comment")) {
							continue;
						}
						if (subchildNode.getLocalName().trim()
								.equals("OperationTarget")) {
							String val = subchildNode.getNodeValue();
							try {
								st.addOperationTarget(val);
							} catch (Exception ex) {
								log.error(
										"Invalid operation target definition detected in filter's service target!"
												+ " Filter Name=" + name
												+ ". Skipping registration! ",
										ex);
								continue;
							}
						}
					}
					def.addServiceTarget(st);

				} else if (aChildNode.getLocalName().trim()
						.equals("AdviceReference")) {
					NamedNodeMap subChildAttributes = aChildNode
							.getAttributes();
					// Assign custom property value
					// Read & set attribute values
					String className = null;
					for (int j = 0; j < subChildAttributes.getLength(); j++) {
						Node subchildAttribute = subChildAttributes.item(j);
						if (subchildAttribute == null
								|| subchildAttribute.getNodeValue() == null
								|| subchildAttribute.getNodeValue().trim()
										.length() == 0) {
							continue;
						}
						String val = subchildAttribute.getNodeValue().trim();
						if (subchildAttribute.getLocalName()
								.equals("className")) {
							className = val;
							break;
						}
					}
					if (className == null) {
						log.error("Invalid Advice reference detected in filter!"
								+ " Filter Name="
								+ name
								+ ". Skipping registration! ");
						continue;
					}
					try {
						Class<?> clazz = Class.forName(className);
						if (!IExecutionAdvice.class.isAssignableFrom(clazz)) {
							log.error("Invalid Advice reference detected in filter!"
									+ " Filter Name="
									+ name
									+ ". Skipping registration! ");
							continue;
						}

						IExecutionAdvice advice = (IExecutionAdvice) clazz
								.newInstance();
						def.addAdvice(advice);
					} catch (Exception ex) {
						log.error(
								"Invalid Advice reference detected in filter!"
										+ " Filter Name=" + name
										+ ". Skipping registration! ", ex);
						continue;
					}
				}
			}

			if (def.getServiceTargetDefinitions() == null
					|| def.getServiceTargetDefinitions().size() <= 0
					|| def.getAdvice().size() <= 0) {
				log.warn("Invalid filter definition!  Skipping definition. "
						+ def.getName());
				continue;
			}
			try {
				Registry.registerFilter(def);
			} catch (Exception e) {
				log.warn("Invalid Filter definition!  Skipping definition. "
						+ def.getName(), e);
			}
		}
	}

	/**
	 * Registers defined services to Registry
	 */
	private static void registerServicesFromConfig(Document XMLDoc) {
		log.info("XMLConfigHelper registering Services ...");
		// Find Components
		// NodeList nodes = XMLDoc.getElementsByTagName("Service");
		NodeList nodes = XMLDoc.getElementsByTagNameNS(namespaceURI, "Service");
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			NamedNodeMap attributes = node.getAttributes();
			if (attributes == null) {
				log.error("Invalid Service definition detected !");
				continue;
			}
			String name = null, description = null;
			Boolean authenticationRequiredToUse = null, authenticationRequiredToInspect = null, lazy = null;
			String locator = null, componentTarget = null, lifeContext = null;

			// Read & set attribute values
			for (int j = 0; j < attributes.getLength(); j++) {
				Node attribute = attributes.item(j);
				if (attribute == null || attribute.getNodeValue() == null
						|| attribute.getNodeValue().trim().length() == 0) {
					continue;
				}
				String val = attribute.getNodeValue().trim();
				if (attribute.getLocalName().equals("name")) {
					name = val;
				} else if (attribute.getLocalName().equals(
						"authenticationRequiredToUse")) {
					authenticationRequiredToUse = Boolean.valueOf(val);
				} else if (attribute.getLocalName().equals(
						"authenticationRequiredToInspect")) {
					authenticationRequiredToInspect = Boolean.valueOf(val);
				} else if (attribute.getLocalName().equals("description")) {
					description = val;
				} else if (attribute.getLocalName().equals("lazy")) {
					lazy = Boolean.valueOf(val);
				} else if (attribute.getLocalName().equals("locator")) {
					locator = val;
				} else if (attribute.getLocalName().equals("target")) {
					componentTarget = val;
				} else if (attribute.getLocalName().equals("scope")) {
					lifeContext = val;
				}
			}

			// Create definition
			ServiceDefinition def = new ServiceDefinition(name, componentTarget);

			if (authenticationRequiredToInspect != null)
				def.setAuthenticationRequiredToInspect(authenticationRequiredToInspect);
			if (authenticationRequiredToUse != null)
				def.setAuthenticationRequiredToUse(authenticationRequiredToUse);
			if (lazy != null)
				def.setLazy(lazy);
			if (locator != null)
				def.setLocator(locator);
			if (componentTarget != null)
				def.setComponentTarget(componentTarget);
			if (lifeContext != null) {
				try {
					def.setServiceScope(ServiceDefinition.ServiceScope
							.valueOf(lifeContext));
				} catch (Exception ex) {
					log.error(
							"An error occured while setting the life context.The default life context will be used for "
									+ name, ex);
				}
			}
			def.setDescription(description);
			// Child Nodes
			NodeList serviceChilds = node.getChildNodes();

			for (int k = 0; k < serviceChilds.getLength(); k++) {
				Node aChildNode = serviceChilds.item(k);
				if (aChildNode.getLocalName() == null
						|| aChildNode.getLocalName().trim().equals("#text")
						|| aChildNode.getLocalName().trim().equals("#comment")) {
					continue;
				}
				NamedNodeMap childAttributes = aChildNode.getAttributes();
				if (aChildNode.getLocalName().trim()
						.equals("AuthorizedRoleToInspect")) {
					String role = aChildNode.getFirstChild() == null ? ""
							: aChildNode.getFirstChild().getNodeValue();
					def.addAuthorizedRoleToInspect(role.trim());
				} else if (aChildNode.getLocalName().trim()
						.equals("AuthorizedRoleToUse")) {
					String role = aChildNode.getFirstChild() == null ? ""
							: aChildNode.getFirstChild().getNodeValue();
					def.addAuthorizedRoleToUse(role.trim());
				} else if (aChildNode.getLocalName().trim().equals("Operation")) {
					OperationDefinition operation;
					try {
						operation = nodeToOperationMD(aChildNode);
						def.addOperation(operation);
					} catch (GateException e) {
						log.error(e.getMessage(), e);
					}

				} else if (aChildNode.getLocalName().trim().equals("Property")) {
					// Assign custom property value
					// Read & set attribute values
					String key = null;
					String value = null;
					for (int j = 0; j < childAttributes.getLength(); j++) {
						Node childAttribute = childAttributes.item(j);
						if (childAttribute == null
								|| childAttribute.getNodeValue() == null
								|| childAttribute.getNodeValue().trim()
										.length() == 0) {
							continue;
						}
						String val = childAttribute.getNodeValue().trim();
						if (childAttribute.getLocalName().equals("key")) {
							key = val;
						} else if (childAttribute.getLocalName()
								.equals("value")) {
							value = val;
						}
					}
					if (key == null || value == null) {
						log.error("Invalid property definition." + " service ="
								+ def.getName());
						break;
					}
					if (def.containsProperty(key)) {
						log.error("Duplicate property definition."
								+ " service =" + def.getName() + ", property="
								+ key);
						continue;
					}
					def.addProperty(key, value);

				}

			}
			// Registering service
			try {
				Registry.registerService(def);
			} catch (GateException e) {
				log.warn("Invalid service definition." + " Service Name="
						+ def.getName() + ".Skipping service definition...");
			}

		}
	}

	private static OperationDefinition nodeToOperationMD(Node node)
			throws GateException {
		NamedNodeMap attributes = node.getAttributes();
		if (attributes == null) {
			throw new GateException("Operation attributes are not defined!");
		}
		String name = null, description = null;
		String opTarget = null;
		Boolean authenticationRequiredToUse = null, authenticationRequiredToInspect = null;
		Boolean secureProtocolRequiredToInspect = null, secureProtocolRequiredToUse = null, remoteAccessEnabled = null;

		// Read & set attribute values
		for (int j = 0; j < attributes.getLength(); j++) {
			Node attribute = attributes.item(j);
			if (attribute == null || attribute.getNodeValue() == null
					|| attribute.getNodeValue().trim().length() == 0) {
				continue;
			}
			String val = attribute.getNodeValue().trim();
			if (attribute.getLocalName().equals("name")) {
				name = val;
			} else if (attribute.getLocalName().equals(
					"authenticationRequiredToUse")) {
				authenticationRequiredToUse = Boolean.valueOf(val);
			} else if (attribute.getLocalName().equals(
					"authenticationRequiredToInspect")) {
				authenticationRequiredToInspect = Boolean.valueOf(val);
			} else if (attribute.getLocalName().equals("description")) {
				description = val;
			} else if (attribute.getLocalName().equals("methodTarget")) {
				opTarget = val;
			} else if (attribute.getLocalName().equals(
					"secureProtocolRequiredToInspect")) {
				secureProtocolRequiredToInspect = Boolean.valueOf(val);
			} else if (attribute.getLocalName().equals(
					"secureProtocolRequiredToUse")) {
				secureProtocolRequiredToUse = Boolean.valueOf(val);
			} else if (attribute.getLocalName().equals("remoteAccessEnabled")) {
				remoteAccessEnabled = Boolean.valueOf(val);
			}
		}
		OperationDefinition def = new OperationDefinition(name, opTarget,
				description);
		if (authenticationRequiredToInspect != null) {
			def.setAuthenticationRequiredToInspect(authenticationRequiredToInspect);
		}
		if (authenticationRequiredToUse != null) {
			def.setAuthenticationRequiredToUse(authenticationRequiredToUse);
		}
		if (secureProtocolRequiredToInspect != null) {
			def.setSecureProtocolRequiredToInspect(secureProtocolRequiredToInspect);
		}
		if (secureProtocolRequiredToUse != null) {
			def.setSecureProtocolRequiredToUse(secureProtocolRequiredToUse);
		}
		if (remoteAccessEnabled != null) {
			def.setRemoteAccessEnabled(remoteAccessEnabled);
		}
		// Child Nodes
		NodeList childNodes = node.getChildNodes();
		for (int k = 0; k < childNodes.getLength(); k++) {
			Node aChildNode = childNodes.item(k);
			if (aChildNode.getLocalName() == null
					|| aChildNode.getLocalName().trim().equals("#text")
					|| aChildNode.getLocalName().trim().equals("#comment")) {
				continue;
			}
			if (aChildNode.getLocalName().trim()
					.equals("AuthorizedRoleToInspect")) {
				String role = aChildNode.getFirstChild() != null ? aChildNode
						.getFirstChild().getNodeValue() : null;
				if (role != null)
					def.addAuthorizedRoleToInspect(role.trim());
			} else if (aChildNode.getLocalName().trim()
					.equals("AuthorizedRoleToUse")) {
				String role = aChildNode.getFirstChild() != null ? aChildNode
						.getFirstChild().getNodeValue() : null;
				if (role != null)
					def.addAuthorizedRoleToUse(role.trim());
			}
		}
		return def;
	}

	/**
	 * Creates and returns new DocumentBuilder object
	 * 
	 * @return DocumentBuilder
	 * @throws ParserConfigurationException
	 */
	private static DocumentBuilder getDocumentBuilder()
			throws ParserConfigurationException {
		// Step 1: create a DocumentBuilderFactory and configure it
		DocumentBuilderFactory lOBdbf = DocumentBuilderFactory.newInstance();
		// Set various configuration options
		lOBdbf.setValidating(false);
		lOBdbf.setIgnoringComments(false);
		lOBdbf.setIgnoringElementContentWhitespace(false);
		lOBdbf.setCoalescing(false);
		// The opposite of creating entity ref nodes is expanding them inline
		lOBdbf.setExpandEntityReferences(true);
		lOBdbf.setNamespaceAware(true);
		// At this point the DocumentBuilderFactory instance can be saved
		// and reused to create any number of DocumentBuilder instances
		// with the same configuration options.

		// Step 2: create a DocumentBuilder that satisfies the constraints
		// specified by the DocumentBuilderFactory
		DocumentBuilder lOBDocumentBuilder = lOBdbf.newDocumentBuilder();

		return lOBDocumentBuilder;
	}

	private static void lockState() {
		lock.lock();
	}

	private static void releaseState() {
		if (lock.isHeldByCurrentThread())
			lock.unlock();
	}
}
