package net.sf.corn.gate.jsonrpc;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import javax.imageio.ImageIO;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.sf.corn.converter.json.JsTypeObject;
import net.sf.corn.converter.json.JsonConverter;
import net.sf.corn.gate.CallDelegator;
import net.sf.corn.gate.GateException;
import net.sf.corn.gate.IClientContext;
import net.sf.corn.gate.OperationDefinition;
import net.sf.corn.gate.Registry;
import net.sf.corn.gate.ServiceDefinition;
import net.sf.corn.gate.ServiceRequest;
import net.sf.corn.gate.ServiceResponse;
import net.sf.corn.gate.web.WebClientContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This servlet is a dispatcher implementation for JSON-RPC and HTTP (get&post) type requests. 
 * This servletsupports requests-response interaction pattern, so clients of this servlet
 * need to wait for their responses.You can map this servlet to both secure and
 * non-secure urls.
 * 
 * @author Serhat Dirik
 * 
 */
@WebServlet(name="JsonRpcDispatcher",displayName="JsonRpcDispatcher",asyncSupported=false,loadOnStartup=1,urlPatterns={"/jsonrpc/*"})
public class JsonRpcDispatcher extends HttpServlet {

	private static final long serialVersionUID = -4475649233825151185L;
	private static Logger log = LoggerFactory.getLogger(JsonRpcDispatcher.class);
	private static Map<String, Object> resources = new ConcurrentHashMap<String, Object>();
	private final static int buf_size = 4096;
	private static boolean keepalive = false;
	private static boolean showRegistry=true;
	// Constants
	private static final String C_UTF = "UTF-8";
	private static final String C_DEFAULT_CONTENT = "text/plain;charset=utf-8";
	// private static final String C_HTML_CONTENT = "text/html";
	private static final String C_CONNECTION_KEY = "Connection";
	private static final String C_CONNECTION_VAL_CLOSE = "close";
	private static final String C_CONNECTION_VAL_KEEP_ALIVE = "keep-alive";
	private static final String C_CACHE_CONTROL_KEY = "Cache-Control";
	private static final String C_CACHE_CONTROL_VAL = "no-cache";
	private static final String C_EXPIRES_KEY = "Expires";
	private static final String C_EXPIRES_VAL = " Thu, 01 Dec 1974 16:00:00 GMT";

	// private static final String C_KEEPALIVE_VAL = "keep-alive";
	private static final String C_HEADER = "header";
	private static final String C_PARAMS = "params";
	private static final String C_NULL = "null";
	private static final String C_SERVICE = "service";
	private static final String C_OPERATION = "operation";
	private static final String C_ID = "id";
	private static final String C_COMMA = ",";
	private static final String C_BRACKET_OPEN = "[";
	private static final String C_BRACKET_CLOSE = "]";
	private static final String C_QUOTA = "\"";
	private static final String C_CBRACKET_OPEN = "{";
	private static final String C_CBRACKET_CLOSE = "}";
	private static final String C_COLUMN = ":";
	private static final String C_SLASH = "/";
	private static final String C_DOT = ".";
	private static final String C_RESOURCE_PARAMETER = "resource";
	private static final String C_SERVICES_PARAMETER = "services";
	private static final String C_CHECK_PARAMETER = "check";
	private static final String LAST_REQUEST = "JSONRPC_LastRequest";
	private static final String LAST_RESPONSE = "JSONRPC_LastResponse";
	@Override
	public void init(ServletConfig config) {
		log.info("Initializing JsonRpcDispatcher...");
		if (config.getInitParameter("keepalive") != null
				&& config.getInitParameter("keepalive").equals("0")) {
			keepalive = false;
		}
		log.info("keepalive=" + keepalive);
		if (config.getInitParameter("showRegistry") != null
				&& ( config.getInitParameter("showRegistry").equals("0")|| config.getInitParameter("showRegistry").trim().equalsIgnoreCase("false") )) {
			showRegistry = false;
		}
		log.info("showRegistry=" + showRegistry);
	}

	@Override
	protected void doGet(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		doGetPost(request, response, true);
	}

	@Override
	protected void doPost(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		doGetPost(request, response, false);
	}

	private void doGetPost(HttpServletRequest httpRequest,
			HttpServletResponse httpResponse, boolean isGet)
			throws ServletException, IOException {

		if (log.isDebugEnabled()) {
			log.debug("method :" + httpRequest.getMethod() + ", isGet:" + isGet);
		}
		// The requested is only the servlet url
		if (httpRequest.getPathInfo() == null) {
			if (isGet) {
				if (httpRequest.getParameter(C_RESOURCE_PARAMETER) != null) {
					if (log.isDebugEnabled()) {
						log.debug(httpRequest.getParameter(C_RESOURCE_PARAMETER)+" requested");
					}
					sendRequestedResource(httpRequest, httpResponse);
					return;
				} else if (httpRequest.getParameter(C_CHECK_PARAMETER) != null) {
					if (log.isDebugEnabled()) {
						log.debug("ping request!");
					}
					sendCheckResponse(httpRequest, httpResponse);
					return;
				} else if (httpRequest.getParameter(C_SERVICES_PARAMETER) != null && showRegistry) {
					if (log.isDebugEnabled()) {
						log.debug("service list requested");
					}
					sendServiceList(httpRequest, httpResponse);
					return;
				}

				// Http get message must contain service and the op name
				httpResponse
						.sendError(
								HttpServletResponse.SC_BAD_REQUEST,
								" The requested url must contain a service target and method name for Http/Get.");
				return;
			}

		}

		JsTypeObject json_res = processServletRequest(httpRequest, httpResponse,
				isGet);
		if (json_res != null && log.isDebugEnabled()) {
			log.debug("Json-Rpc Response: " + json_res.toString());
		}
		// Write the response
		OutputStream out = httpResponse.getOutputStream();
		byte[] bout = null;
		bout = json_res.toJsonString().getBytes(C_UTF);
		// Encode using UTF-8
		httpResponse.setContentType(C_DEFAULT_CONTENT);
		if (keepalive) {
			httpResponse.setHeader("Keep-Alive", "300");
			httpResponse.setHeader(C_CONNECTION_KEY,
					C_CONNECTION_VAL_KEEP_ALIVE);
		} else {
			httpResponse.setHeader(C_CONNECTION_KEY, C_CONNECTION_VAL_CLOSE);// httpRequest.getHeader(C_KEEPALIVE_KEY));
		}
		httpResponse.setHeader(C_CACHE_CONTROL_KEY, C_CACHE_CONTROL_VAL);
		httpResponse.setHeader(C_EXPIRES_KEY, C_EXPIRES_VAL);
		out.write(bout);
		httpResponse.setContentLength(bout.length);
		out.flush();
		out.close();
	}

	private void sendServiceList(HttpServletRequest httpRequest,
			HttpServletResponse httpResponse) throws ServletException,
			IOException {
		httpResponse.setContentType("text/html");
		OutputStream out = httpResponse.getOutputStream();
		StringBuilder sb = new StringBuilder();
		sb.append("<!doctype html><head><title>Gate Service List</title></head><body><ol>");
		WebClientContext context = new WebClientContext(httpRequest);
		Map<String,ServiceDefinition> services = Registry.getServiceDefinitions(context);
		for (ServiceDefinition sdef : services.values()) {
			sb.append("<li><b style=\"color:red\">").append(sdef.getName()).append("</b>").append(sdef.getDescription()== null ?"":": "+sdef.getDescription()).append("<br>");
			if(sdef.getOperations().size()> 0){
				sb.append("<ul>");
				for(OperationDefinition opdef:sdef.getOperations()){
					sb.append("<li>");
					sb.append("<i style=\"color:Blue\">").append(opdef.getName()).append("</i>").append(opdef.getDescription()== null ?"": ":"+opdef.getDescription()).append("<br>");
					sb.append("<ul><code>");
					try {
						List<Method> methods = Registry.findNameMatchingMethods(sdef.getName(), opdef.getName(),context);
						for (Method method : methods) {
							sb.append("<li style=\"color:DarkBlue\">").append(method.getReturnType().getName()).append(" ");
							sb.append(opdef.getName()).append("(");
							boolean first=true;
							for(Class<?> pclazz:method.getParameterTypes()){
								if(!first) sb.append(",");
								sb.append(pclazz.getName());
								first=false;
							}
							sb.append(")</li>");
						}
					} catch (GateException e) {
					}			
					
					sb.append("</code></ul></li>");		
				}
				sb.append("</ul>");
			}
			
			sb.append("</li>");
		}
		sb.append("</ol></body>");
		
		
		byte[] bout = null;
		bout = sb.toString().getBytes("UTF-8");
		out.write(bout);
		httpResponse.setContentLength(bout.length);
		out.flush();
		out.close();
	}
	
	private void sendCheckResponse(HttpServletRequest httpRequest,
			HttpServletResponse httpResponse) throws ServletException,
			IOException {
		httpResponse.setContentType("text/javascript");
		OutputStream out = httpResponse.getOutputStream();
		String text = "\"OK\"";
		byte[] bout = null;
		bout = text.getBytes("UTF-8");
		out.write(bout);
		httpResponse.setContentLength(bout.length);
		out.flush();
		out.close();
	}

	private void sendRequestedResource(HttpServletRequest httpRequest,
			HttpServletResponse httpResponse) throws ServletException,
			IOException {

		String resource = httpRequest.getParameter(C_RESOURCE_PARAMETER);
		String extention = null;
		boolean isText = false;
		if (resource == null || resource.trim().length() == 0) {
			// Http get message must contain service and the op name
			httpResponse.sendError(HttpServletResponse.SC_BAD_REQUEST,
					" The requested resource " + resource + " is not found!");
			return;
		}

		OutputStream out = httpResponse.getOutputStream();
		if (resource.endsWith(".js")) {
			httpResponse.setContentType("text/javascript");
			extention = "js";
			isText = true;
		} else if (resource.endsWith(".css")) {
			httpResponse.setContentType("text/css");
			extention = "css";
			isText = true;
		} else if (resource.endsWith(".gif")) {
			httpResponse.setContentType("image/gif");
			extention = "gif";
			isText = false;
		} else if (resource.endsWith(".png")) {
			httpResponse.setContentType("image/png");
			extention = "gif";
			isText = false;
		} else if (resource.endsWith(".jpg") || resource.endsWith(".jpeg")) {
			httpResponse.setContentType("image/jpeg");
			extention = "jpg";
			isText = false;
		} else if (resource.endsWith(".xml") || resource.endsWith(".xsl")
				|| resource.endsWith(".xsd")) {
			httpResponse.setContentType("text/xml"); // XSL has to be served as
														// XML.
			extention = "xml";
			isText = true;
		} else if (resource.endsWith(".htm")) {
			httpResponse.setContentType("text/html");
			extention = "htm";
			isText = true;
		} else if (resource.endsWith(".html")) {
			httpResponse.setContentType("text/html");
			extention = "html";
			isText = true;
		} else {
			httpResponse.sendError(HttpServletResponse.SC_NOT_FOUND, resource
					+ " not found.");
			return;
		}

		if (isText) {
			if (resources.containsKey(resource)) {
				String text = (String) resources.get(resource);
				byte[] bout = null;
				bout = text.getBytes("UTF-8");
				out.write(bout);
				httpResponse.setContentLength(bout.length);
				out.flush();
				out.close();
			} else {
				InputStream is = this.getClass().getResourceAsStream(
						resource.toLowerCase()); // context.getResourceAsStream(resource);
				if (is == null) {
					httpResponse.sendError(HttpServletResponse.SC_NOT_FOUND,
							resource + " is not found");
					return;
				}
				InputStreamReader lsInputReader = new InputStreamReader(is);
				BufferedReader lsBufferedReader = new BufferedReader(
						lsInputReader);
				String lsInput = null;
				StringBuilder sb = new StringBuilder();
				int len = 0;
				while ((lsInput = lsBufferedReader.readLine()) != null) {
					byte[] bout = null;
					lsInput += "\n";
					sb.append(lsInput);
					bout = lsInput.getBytes("UTF-8");
					len += bout.length;
					out.write(bout);
				}
				httpResponse.setContentLength(len);
				out.flush();
				out.close();
				resources.put(resource, sb.toString());
			}

		} else {
			if (resources.containsKey(resource)) {
				BufferedImage img = (BufferedImage) resources.get(resource
						.toLowerCase());
				ImageIO.write(img, extention, out);
				out.flush();
				out.close();
			} else {
				InputStream is = this.getClass().getResourceAsStream(resource); // context.getResourceAsStream(resource);
				if (is == null) {
					httpResponse.sendError(HttpServletResponse.SC_NOT_FOUND,
							resource + " is not found");
					return;
				}
				BufferedImage img = ImageIO.read(is);
				ImageIO.write(img, extention, out);
				out.flush();
				out.close();
				resources.put(resource, img);
			}

		}
	}

	private JsTypeObject processServletRequest(HttpServletRequest httpRequest,
			HttpServletResponse httpResponse, boolean isGet)
			throws ServletException, IOException {

		JsTypeObject json_res = null;
		// Detect the charset
		String charset = httpRequest.getCharacterEncoding();
		if (charset == null) {
			charset = C_UTF;
		}
		// Try to find service & op names
		String operationName = null;
		String serviceName = null;

		if (httpRequest.getPathInfo() != null) {
			// If the path info is not null than we need to detect service and
			// operation name from it
			serviceName = httpRequest.getPathInfo().substring(1)
					.replace(C_SLASH, C_DOT);
			int serviceDelim = serviceName.indexOf(C_DOT);

			if (serviceDelim < 0 || serviceDelim >= serviceName.length() - 1
					|| serviceDelim == 0) {
				json_res = new JsonRpcResponse("0",
						"InvalidArgument",
						"There is no operation exist as "
								+ httpRequest.getPathInfo()).toJSON();
				return json_res;

			}
			operationName = serviceName.substring(serviceDelim + 1);
			serviceName = serviceName.substring(0, serviceDelim);
		}

		// Let's check incoming parameters
		String jsonString = null;

		if (!isGet) {
			BufferedReader in = new BufferedReader(new InputStreamReader(
					httpRequest.getInputStream(), charset));
			// Read the request
			CharArrayWriter postedData = new CharArrayWriter();
			char buf[] = new char[buf_size];
			int ret;
			while ((ret = in.read(buf, 0, buf_size)) != -1)
				postedData.write(buf, 0, ret);
			jsonString = postedData.toString();
		}

		if (jsonString == null || !jsonString.contains(C_SERVICE)
				|| !jsonString.contains(C_OPERATION)) {
			if (serviceName == null || operationName == null) {
				json_res = new JsonRpcResponse("0",
						"ParsingException",
						" The requested url must contain a service and an operation name ")
						.toJSON();
				return json_res;
			}
			// Try to get from parameter map

			Map<String, String[]> parameters = null;
			if (jsonString != null) {
				parameters = resolveParameters(charset, jsonString);
			} else {
				parameters = httpRequest.getParameterMap();
			}

			jsonString = convertParamsToJsonString(parameters, serviceName,
					operationName);
			if (jsonString == null) {
				json_res = new JsonRpcResponse("0",
						"InvalidArgument",
						" Incoming parameters couldn't be converted to json data.")
						.toJSON();
				return json_res;
			}
		}
		if (log.isDebugEnabled())
			log.debug(httpRequest.getServletPath() + " recieved: " + jsonString);

		// Try to convert JsonRpcRequest
		JsonRpcRequest jsonReq;
		try {
			jsonReq = new JsonRpcRequest(jsonString);
		} catch (Exception e) {
			log.error("Couldn't create the JsonRpcRequest", e);
			json_res = new JsonRpcResponse("0",
					"ParsingException",
					" Incoming parameters couldn't be converted to json rpc request.")
					.toJSON();
			return json_res;
		}
		if (log.isDebugEnabled()) {
			log.debug("Json-Rpc Request: " + jsonReq.toString());
		}
		// 1- Convert incoming request to JSONObject than JSONRequest
		if (jsonReq.getRequestId() == null) {
			jsonReq.setRequestId(UUID.randomUUID().toString());
			if (log.isDebugEnabled()) {
				log.debug("A new id assigned : " + jsonReq.getRequestId());
			}
		}

		// All responses from now on will be returned as JSON

		JsonRpcResponse rpcRes = processJsonRequest(httpRequest, jsonReq);
		json_res = rpcRes.toJSON();
		return json_res;

	}

	private static Map<String, String[]> resolveParameters(String charset,
			String strData) throws ServletException, IOException {
		Map<String, String[]> parameters = new HashMap<String, String[]>();
		String encodedData = URLDecoder.decode(strData, charset);
		StringTokenizer paramTokenizer = new StringTokenizer(encodedData, "&");
		while (paramTokenizer.hasMoreTokens()) {
			String pairSt = paramTokenizer.nextToken();
			StringTokenizer pairTokenizer = new StringTokenizer(pairSt, "=");
			if (pairTokenizer.countTokens() > 2
					|| pairTokenizer.countTokens() < 1) {
				parameters.clear();
				break;
			}
			if (pairTokenizer.countTokens() == 2)
				parameters.put(pairTokenizer.nextToken(),
						new String[] { pairTokenizer.nextToken() });
		}
		return parameters;
	}

	private String convertParamsToJsonString(Map<String, String[]> parameters,
			String serviceName, String operationName) {
		// Url will be consumed as service name and post data or query string as
		// the rest of message (params id header)
		// if data is consumed from client, than it must contains parameters
		StringBuilder sb = new StringBuilder();
		sb.append(C_CBRACKET_OPEN);
		sb.append(C_QUOTA).append(C_SERVICE).append(C_QUOTA).append(C_COLUMN)
				.append(C_QUOTA).append(serviceName).append(C_QUOTA);
		sb.append(C_COMMA).append(C_QUOTA).append(C_OPERATION).append(C_QUOTA)
				.append(C_COLUMN).append(C_QUOTA).append(operationName)
				.append(C_QUOTA);

		String[] vals = parameters.get(C_ID);
		String val = vals != null && vals.length > 0 ? vals[0] : null;
		if (val != null) {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_ID);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			sb.append(C_QUOTA);
			sb.append(val);
			sb.append(C_QUOTA);
		} else {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_ID);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			sb.append(C_QUOTA);
			sb.append(UUID.randomUUID());
			sb.append(C_QUOTA);
		}
		vals = parameters.get(C_HEADER);
		val = vals != null && vals.length > 0 ? vals[0] : null;
		if (val != null) {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_HEADER);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			if (val.startsWith(C_CBRACKET_OPEN)) {
				sb.append(val);
			} else {
				sb.append(C_CBRACKET_OPEN);
				sb.append(val);
				sb.append(C_CBRACKET_CLOSE);
			}

		} else {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_HEADER);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			sb.append(C_NULL);
		}
		boolean paramsAdded = false;
		vals = parameters.get(C_PARAMS);
		val = vals != null && vals.length > 0 ? vals[0] : null;
		if (val != null) {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_PARAMS);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			sb.append(C_BRACKET_OPEN);
			sb.append(val);
			sb.append(C_BRACKET_CLOSE);
			paramsAdded = true;
		} else {
			// If parameters not specified explicitly than use any parameter as
			// paramater
			for (String key : parameters.keySet()) {
				vals = parameters.get(key);
				val = vals[0];
				key = key.toLowerCase();
				boolean first = true;

				if (!key.equals(C_ID) && !key.equals(C_HEADER)) {
					if (first) {
						sb.append(C_COMMA);
						sb.append(C_QUOTA);
						sb.append(C_PARAMS);
						sb.append(C_QUOTA);
						sb.append(C_COLUMN);
						sb.append(C_BRACKET_OPEN);
					}

					if (!first)
						sb.append(C_COMMA);
					sb.append(val);
					first = false;
					paramsAdded = true;
				}
				if (paramsAdded) {
					sb.append(C_BRACKET_CLOSE);
				}
			}
		}

		if (!paramsAdded) {
			sb.append(C_COMMA);
			sb.append(C_QUOTA);
			sb.append(C_PARAMS);
			sb.append(C_QUOTA);
			sb.append(C_COLUMN);
			sb.append(C_NULL);
		}
		sb.append(C_CBRACKET_CLOSE);
		return sb.toString();
	}

	private JsonRpcResponse processJsonRequest(HttpServletRequest httpRequest,
			JsonRpcRequest jsonReq) throws IOException {
		// Process the request
		JsonRpcResponse json_res = null;
		boolean continueState = true;

		boolean repetitiveCall = false;
		// Idempotency check to prevent repetitive calls
		if (httpRequest.getSession() != null
				&& httpRequest.isRequestedSessionIdValid()) {
			JsonRpcRequest lastReq = (JsonRpcRequest) httpRequest.getSession()
					.getAttribute(LAST_REQUEST);
			if (lastReq != null
					&& lastReq.getRequestId().equals(jsonReq.getRequestId())
					&& lastReq.getServiceName()
							.equals(jsonReq.getServiceName())
					&& lastReq.getOperationName().equals(
							jsonReq.getOperationName())) {
				if (lastReq.getParameters() == null
						&& jsonReq.getParameters() == null) {
					repetitiveCall = true;
				} else if (lastReq.getParameters() != null
						&& jsonReq.getParameters() != null) {
					repetitiveCall = true;
					for (int i = 0; i < lastReq.getParameters().size(); i++) {
						JsTypeObject lastReqObj = lastReq.getParameters().get(i);
						Object reqObj = jsonReq.getParameters().get(i);
						if (!lastReqObj.equals(reqObj)) {
							repetitiveCall = false;
							break;
						}
					}
				}
			}
		}

		if (repetitiveCall) {
			// Return the same result
			json_res = (JsonRpcResponse) httpRequest.getSession().getAttribute(
					LAST_RESPONSE);
			if (json_res != null && log.isDebugEnabled()) {
				log.debug("A call has been made with same id and parameters to the same service and op. Same saved result will be returned");
			}
			continueState = false;
		}

		// 4-Transform Parameters to Java Objects
		// Wrap the request
		IClientContext clientContext = null;
		Object[] parameters = new Object[jsonReq.getParameters().size()];
		if (continueState) {

			WebClientContext webContext = null;
			
			webContext = new WebClientContext(httpRequest);
			Map<String, Object> header = jsonReq.getHeader();
			for (String key : header.keySet()) {
				webContext.addToRequestHeader(key, header.get(key));
			}
			clientContext = webContext;
 
			// At this point we must be sure that we know that the client
			// requested
			// a service and a method
			// Check if the service and operations are exist
			String serviceName = jsonReq.getServiceName();
			String operationName = jsonReq.getOperationName();

			ServiceDefinition serviceDef = null;
			OperationDefinition operationDef = null;
			JsonConverter converter = new JsonConverter();
			Method targetMethod = null;
			List<Method> methodCandidates = null;
			try {
				serviceDef = Registry.getServiceDefinition(serviceName,
						clientContext);
				operationDef = serviceDef.getOperation(operationName);
				if (operationDef == null) {
					throw new GateException("The requested operation " + serviceName
							+ "." + operationName + " is not found!")
							.code(GateException.CODE_REGISTRY_OPERATION_NOT_FOUND);
				}

				methodCandidates = Registry.findNameMatchingMethods(
						jsonReq.getServiceName(), jsonReq.getOperationName(),
						clientContext);

				for (Method candidate : methodCandidates) {
					if (jsonReq.getParameters() == null
							&& candidate.getParameterTypes().length > 0)
						continue;
					if (candidate.getParameterTypes().length != jsonReq
							.getParameters().size())
						continue;

					boolean allMatched = true;
					for (int i = 0; i < parameters.length; i++) {
						try {
							Class<?> paramClass = candidate.getParameterTypes()[i];
							JsTypeObject source = jsonReq.getParameters().get(i);
							Class<?> sourceClass = source.getJavaClass();
							if(sourceClass != null && ! paramClass.isAssignableFrom(sourceClass)){
								allMatched = false;
								break;
							}
							if (!converter.isConvertible(paramClass, source)) {
								allMatched = false;
								break;
							}
							paramClass = sourceClass != null? sourceClass : paramClass;
							
							parameters[i] = converter
									.toJava(paramClass, source);
						} catch (Exception e) {
							allMatched = false;
							log.warn(
									"An exception occured while type  conversion",
									e);
							break;
						}
					}

					if (allMatched) {
						targetMethod = candidate;
						break;
					}
					if (targetMethod != null)
						break;
				}

				if (targetMethod == null) {
					continueState = false;
					log.error("Service " + jsonReq.getServiceName()
							+ " does not contain the method "
							+ jsonReq.getOperationName()
							+ " that matched to the requested signature");
					json_res = new JsonRpcResponse(
							jsonReq.getRequestId(),
							"OperationNotFound",
							"Service "
									+ jsonReq.getServiceName()
									+ " does not contain the method "
									+ jsonReq.getOperationName()
									+ " that matched to the requested signature");

				} else {
					if (log.isDebugEnabled()) {
						log.debug("Request:" + jsonReq.getRequestId()
								+ ":Method signature is matched "
								+ targetMethod.toString());
					}
				}
			} catch (GateException e) {
				continueState = false;
				log.error(
						"An error occured while trying to access Service "
								+ jsonReq.getServiceName() + "."
								+ jsonReq.getOperationName(), e);
				json_res = new JsonRpcResponse(jsonReq.getRequestId(), "GateException_e.getCode()",
						"An error occured while trying to access Service "
								+ jsonReq.getServiceName() + "."
								+ jsonReq.getOperationName() + "\n"
								+ e.getMessage());
			}

		}

		// 4- Made the actual call
		if (continueState && json_res == null) {

			try {

				ServiceRequest serviceRequest = new ServiceRequest(
						jsonReq.getServiceName(), jsonReq.getOperationName(),
						parameters, clientContext);
				Map<String, Object> header = jsonReq.getHeader();
				for (String key : header.keySet()) {
					serviceRequest.addToHeader(key, header.get(key));
				}

				ServiceResponse executionResponse = CallDelegator
						.execute(serviceRequest);
				if (executionResponse.hasResult()) {
					try {
						JsonConverter converter = new JsonConverter();
						json_res = new JsonRpcResponse(jsonReq.getRequestId(),
								converter.toTargetType(executionResponse
										.getReturnedValue()));
					} catch (Exception e) {
						continueState = false;
						log.error(
								"An error occured while transforming the result of the service operation '"
										+ jsonReq.getServiceName() + "."
										+ jsonReq.getOperationName(), e);
						json_res = new JsonRpcResponse(jsonReq.getRequestId(),
								e.getClass().getName(),
								"An error occured while transforming the result of the service operation '"
										+ jsonReq.getServiceName() + "."
										+ jsonReq.getOperationName() + "' :"
										+ e.getMessage());

					}
				} else {
					if (log.isDebugEnabled()) {
						log.debug(
								"Service operation '"
										+ jsonReq.getServiceName() + "."
										+ jsonReq.getOperationName()
										+ "' threw an exception",
								executionResponse.getReturnedException());
					}

					json_res = new JsonRpcResponse(jsonReq.getRequestId(),
							executionResponse.getReturnedException().getClass()
									.getName(), "Service operation '"
									+ jsonReq.getServiceName()
									+ "."
									+ jsonReq.getOperationName()
									+ "' threw an exception: "
									+ executionResponse.getReturnedException()
											.getMessage());
				}
				// Put header values from the context
				for (String key : clientContext.responseHeaderKeySet()) {
					json_res.putToHeader(key,
							clientContext.getFromResponseHeader(key));
				}
			} catch (Exception e) {
				continueState = false;
				log.error("Service operation '" + jsonReq.getServiceName()
						+ "." + jsonReq.getOperationName()
						+ "' threw an unexpected exception", e);
				json_res = new JsonRpcResponse(jsonReq.getRequestId(), e
						.getClass().getName(), "Service operation '"
						+ jsonReq.getServiceName() + "."
						+ jsonReq.getOperationName()
						+ "' threw an unexpected exception: " + e.getMessage());

			}

		}
		// 5- Store request response for Idempotency check
		if (continueState && httpRequest.getSession() != null) {
			// && httpRequest.isRequestedSessionIdValid()) {
			httpRequest.getSession().setAttribute(LAST_REQUEST, jsonReq);
			httpRequest.getSession().setAttribute(LAST_RESPONSE, json_res);
		}

		return json_res;
	}
}
